@@ -7,6 +7,9 @@ import (
 	"github.com/golang/glog"
 	"github.com/openshift/library-go/pkg/operator/events"
 	"github.com/openshift/library-go/pkg/operator/resource/resourceapply"
+	"github.com/openshift/library-go/pkg/operator/resource/resourcemerge"
+	machinecontroller "github.com/openshift/machine-api-operator/pkg/controller/machine"
+	"github.com/openshift/machine-api-operator/pkg/util/conditions"
 	appsv1 "k8s.io/api/apps/v1"
 	corev1 "k8s.io/api/core/v1"
 	apierrors "k8s.io/apimachinery/pkg/api/errors"
@@ -17,8 +20,12 @@ import (
 )
 
 const (
-	deploymentRolloutPollInterval = time.Second
-	deploymentRolloutTimeout      = 5 * time.Minute
+	deploymentRolloutPollInterval     = time.Second
+	deploymentRolloutTimeout          = 5 * time.Minute
+	deploymentMinimumAvailabilityTime = 3 * time.Minute
+	daemonsetRolloutPollInterval      = time.Second
+	daemonsetRolloutTimeout           = 5 * time.Minute
+	machineAPITerminationHandler      = "machine-api-termination-handler"
 )
 
 func (optr *Operator) syncAll(config *OperatorConfig) error {
@@ -26,19 +33,27 @@ func (optr *Operator) syncAll(config *OperatorConfig) error {
 		glog.Errorf("Error syncing ClusterOperatorStatus: %v", err)
 		return fmt.Errorf("error syncing ClusterOperatorStatus: %v", err)
 	}
-	if config.Controllers.Provider != clusterAPIControllerNoOp {
-		if err := optr.syncClusterAPIController(config); err != nil {
-			if err := optr.statusDegraded(err.Error()); err != nil {
-				// Just log the error here.  We still want to
-				// return the outer error.
-				glog.Errorf("Error syncing ClusterOperatorStatus: %v", err)
-			}
-			glog.Errorf("Error syncing machine-api-controller: %v", err)
-			return err
+
+	if config.Controllers.Provider == clusterAPIControllerNoOp {
+		glog.V(3).Info("Provider is NoOp, skipping synchronisation")
+		if err := optr.statusAvailable(); err != nil {
+			glog.Errorf("Error syncing ClusterOperatorStatus: %v", err)
+			return fmt.Errorf("error syncing ClusterOperatorStatus: %v", err)
 		}
-		glog.V(3).Info("Synced up all machine-api-controller components")
+		return nil
 	}
 
+	if err := optr.syncClusterAPIController(config); err != nil {
+		if err := optr.statusDegraded(err.Error()); err != nil {
+			// Just log the error here.  We still want to
+			// return the outer error.
+			glog.Errorf("Error syncing ClusterOperatorStatus: %v", err)
+		}
+		glog.Errorf("Error syncing machine-api-controller: %v", err)
+		return err
+	}
+	glog.V(3).Info("Synced up all machine-api-controller components")
+
 	// In addition, if the Provider is BareMetal, then bring up
 	// the baremetal-operator pod
 	if config.BaremetalControllers.BaremetalOperator != "" {
@@ -63,24 +78,40 @@ func (optr *Operator) syncAll(config *OperatorConfig) error {
 
 func (optr *Operator) syncClusterAPIController(config *OperatorConfig) error {
 	controllersDeployment := newDeployment(config, nil)
-	d, err := optr.deployLister.Deployments(controllersDeployment.GetNamespace()).Get(controllersDeployment.GetName())
-	generation := int64(-1)
-	if err != nil {
-		if !apierrors.IsNotFound(err) {
-			return err
-		}
-	}
-	if d != nil {
-		generation = d.Status.ObservedGeneration
-	}
-
-	_, updated, err := resourceapply.ApplyDeployment(optr.kubeClient.AppsV1(),
-		events.NewLoggingEventRecorder(optr.name), controllersDeployment, generation)
+	expectedGeneration := resourcemerge.ExpectedDeploymentGeneration(controllersDeployment, optr.generations)
+	d, updated, err := resourceapply.ApplyDeployment(optr.kubeClient.AppsV1(),
+		events.NewLoggingEventRecorder(optr.name), controllersDeployment, expectedGeneration)
 	if err != nil {
 		return err
 	}
 	if updated {
-		return optr.waitForDeploymentRollout(controllersDeployment)
+		resourcemerge.SetDeploymentGeneration(&optr.generations, d)
+		err := optr.waitForDeploymentRollout(controllersDeployment, deploymentRolloutPollInterval, deploymentRolloutTimeout)
+		if err != nil {
+			return err
+		}
+	}
+
+	// Sync Termination Handler DaemonSet if supported
+	if config.Controllers.TerminationHandler != clusterAPIControllerNoOp {
+		if err := optr.syncTerminationHandler(config); err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+func (optr *Operator) syncTerminationHandler(config *OperatorConfig) error {
+	terminationDaemonSet := newTerminationDaemonSet(config)
+	expectedGeneration := resourcemerge.ExpectedDaemonSetGeneration(terminationDaemonSet, optr.generations)
+	ds, updated, err := resourceapply.ApplyDaemonSet(optr.kubeClient.AppsV1(),
+		events.NewLoggingEventRecorder(optr.name), terminationDaemonSet, expectedGeneration)
+	if err != nil {
+		return err
+	}
+	if updated {
+		resourcemerge.SetDaemonSetGeneration(&optr.generations, ds)
+		return optr.waitForDaemonSetRollout(terminationDaemonSet)
 	}
 	return nil
 }
@@ -99,34 +130,27 @@ func (optr *Operator) syncBaremetalControllers(config *OperatorConfig) error {
 	}
 
 	metal3Deployment := newMetal3Deployment(config, baremetalProvisioningConfig)
-	generation := int64(-1)
-	d, err := optr.deployLister.Deployments(metal3Deployment.GetNamespace()).Get(metal3Deployment.GetName())
-	if err != nil {
-		if !apierrors.IsNotFound(err) {
-			return err
-		}
-	}
-	if d != nil {
-		generation = d.Status.ObservedGeneration
-	}
-
-	_, updated, err := resourceapply.ApplyDeployment(optr.kubeClient.AppsV1(),
-		events.NewLoggingEventRecorder(optr.name), metal3Deployment, generation)
+	expectedGeneration := resourcemerge.ExpectedDeploymentGeneration(metal3Deployment, optr.generations)
+	d, updated, err := resourceapply.ApplyDeployment(optr.kubeClient.AppsV1(),
+		events.NewLoggingEventRecorder(optr.name), metal3Deployment, expectedGeneration)
 	if err != nil {
 		return err
 	}
 	if updated {
-		return optr.waitForDeploymentRollout(metal3Deployment)
+		resourcemerge.SetDeploymentGeneration(&optr.generations, d)
+		return optr.waitForDeploymentRollout(metal3Deployment, deploymentRolloutPollInterval, deploymentRolloutTimeout)
 	}
 
 	return nil
 }
 
-func (optr *Operator) waitForDeploymentRollout(resource *appsv1.Deployment) error {
+func (optr *Operator) waitForDeploymentRollout(resource *appsv1.Deployment, pollInterval, rolloutTimeout time.Duration) error {
 	var lastError error
-	err := wait.Poll(deploymentRolloutPollInterval, deploymentRolloutTimeout, func() (bool, error) {
+	err := wait.Poll(pollInterval, rolloutTimeout, func() (bool, error) {
 		d, err := optr.deployLister.Deployments(resource.Namespace).Get(resource.Name)
 		if apierrors.IsNotFound(err) {
+			lastError = fmt.Errorf("deployment %s is not found", d.GetName())
+			glog.Error(lastError)
 			return false, nil
 		}
 		if err != nil {
@@ -143,9 +167,27 @@ func (optr *Operator) waitForDeploymentRollout(resource *appsv1.Deployment) erro
 		}
 
 		if d.Generation <= d.Status.ObservedGeneration && d.Status.UpdatedReplicas == d.Status.Replicas && d.Status.UnavailableReplicas == 0 {
+			c := conditions.GetDeploymentCondition(d, appsv1.DeploymentAvailable)
+			if c == nil {
+				lastError = fmt.Errorf("deployment %s is not reporting available yet", d.GetName())
+				glog.V(4).Info(lastError)
+				return false, nil
+			}
+			if c.Status == corev1.ConditionFalse {
+				lastError = fmt.Errorf("deployment %s is reporting available=false", d.GetName())
+				glog.V(4).Info(lastError)
+				return false, nil
+			}
+			if c.LastTransitionTime.Time.Add(deploymentMinimumAvailabilityTime).After(time.Now()) {
+				lastError = fmt.Errorf("deployment %s has been available for less than 3 min", d.GetName())
+				glog.V(4).Info(lastError)
+				return false, nil
+			}
+
 			lastError = nil
 			return true, nil
 		}
+
 		lastError = fmt.Errorf("deployment %s is not ready. status: (replicas: %d, updated: %d, ready: %d, unavailable: %d)", d.Name, d.Status.Replicas, d.Status.UpdatedReplicas, d.Status.ReadyReplicas, d.Status.UnavailableReplicas)
 		glog.V(4).Info(lastError)
 		return false, nil
@@ -156,6 +198,40 @@ func (optr *Operator) waitForDeploymentRollout(resource *appsv1.Deployment) erro
 	return err
 }
 
+func (optr *Operator) waitForDaemonSetRollout(resource *appsv1.DaemonSet) error {
+	var lastError error
+	err := wait.Poll(daemonsetRolloutPollInterval, daemonsetRolloutTimeout, func() (bool, error) {
+		d, err := optr.daemonsetLister.DaemonSets(resource.Namespace).Get(resource.Name)
+		if apierrors.IsNotFound(err) {
+			return false, nil
+		}
+		if err != nil {
+			// Do not return error here, as we could be updating the API Server itself, in which case we
+			// want to continue waiting.
+			lastError = fmt.Errorf("getting DaemonSet %s during rollout: %v", resource.Name, err)
+			glog.Error(lastError)
+			return false, nil
+		}
+
+		if d.DeletionTimestamp != nil {
+			lastError = nil
+			return false, fmt.Errorf("daemonset %s is being deleted", resource.Name)
+		}
+
+		if d.Generation <= d.Status.ObservedGeneration && d.Status.UpdatedNumberScheduled == d.Status.DesiredNumberScheduled && d.Status.NumberUnavailable == 0 {
+			lastError = nil
+			return true, nil
+		}
+		lastError = fmt.Errorf("daemonset %s is not ready. status: (desired: %d, updated: %d, available: %d, unavailable: %d)", d.Name, d.Status.DesiredNumberScheduled, d.Status.UpdatedNumberScheduled, d.Status.NumberAvailable, d.Status.NumberUnavailable)
+		glog.V(4).Info(lastError)
+		return false, nil
+	})
+	if lastError != nil {
+		return lastError
+	}
+	return err
+}
+
 func newDeployment(config *OperatorConfig, features map[string]bool) *appsv1.Deployment {
 	replicas := int32(1)
 	template := newPodTemplateSpec(config, features)
@@ -242,9 +318,9 @@ func newContainers(config *OperatorConfig, features map[string]bool) []corev1.Co
 
 	containers := []corev1.Container{
 		{
-			Name:      "controller-manager",
-			Image:     config.Controllers.Provider,
-			Command:   []string{"/manager"},
+			Name:      "machineset-controller",
+			Image:     config.Controllers.MachineSet,
+			Command:   []string{"/machineset-controller"},
 			Args:      args,
 			Resources: resources,
 		},
@@ -282,3 +358,85 @@ func newContainers(config *OperatorConfig, features map[string]bool) []corev1.Co
 	}
 	return containers
 }
+
+func newTerminationDaemonSet(config *OperatorConfig) *appsv1.DaemonSet {
+	template := newTerminationPodTemplateSpec(config)
+
+	return &appsv1.DaemonSet{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      machineAPITerminationHandler,
+			Namespace: config.TargetNamespace,
+			Annotations: map[string]string{
+				maoOwnedAnnotation: "",
+			},
+			Labels: map[string]string{
+				"api":     "clusterapi",
+				"k8s-app": "termination-handler",
+			},
+		},
+		Spec: appsv1.DaemonSetSpec{
+			Selector: &metav1.LabelSelector{
+				MatchLabels: map[string]string{
+					"api":     "clusterapi",
+					"k8s-app": "termination-handler",
+				},
+			},
+			Template: *template,
+		},
+	}
+}
+
+func newTerminationPodTemplateSpec(config *OperatorConfig) *corev1.PodTemplateSpec {
+	containers := newTerminationContainers(config)
+
+	return &corev1.PodTemplateSpec{
+		ObjectMeta: metav1.ObjectMeta{
+			Labels: map[string]string{
+				"api":     "clusterapi",
+				"k8s-app": "termination-handler",
+			},
+		},
+		Spec: corev1.PodSpec{
+			Containers:         containers,
+			PriorityClassName:  "system-node-critical",
+			NodeSelector:       map[string]string{machinecontroller.MachineInterruptibleInstanceLabelName: ""},
+			ServiceAccountName: machineAPITerminationHandler,
+			HostNetwork:        true,
+		},
+	}
+}
+
+func newTerminationContainers(config *OperatorConfig) []corev1.Container {
+	resources := corev1.ResourceRequirements{
+		Requests: map[corev1.ResourceName]resource.Quantity{
+			corev1.ResourceMemory: resource.MustParse("20Mi"),
+			corev1.ResourceCPU:    resource.MustParse("10m"),
+		},
+	}
+	terminationArgs := []string{
+		"--logtostderr=true",
+		"--v=3",
+		"--node-name=$(NODE_NAME)",
+		fmt.Sprintf("--namespace=%s", config.TargetNamespace),
+		"--poll-interval-seconds=5",
+	}
+	return []corev1.Container{
+		{
+			Name:      "termination-handler",
+			Image:     config.Controllers.TerminationHandler,
+			Command:   []string{"/termination-handler"},
+			Args:      terminationArgs,
+			Resources: resources,
+			Env: []corev1.EnvVar{
+				{
+					Name: "NODE_NAME",
+					ValueFrom: &corev1.EnvVarSource{
+						FieldRef: &corev1.ObjectFieldSelector{
+							FieldPath: "spec.nodeName",
+						},
+					},
+				},
+			},
+		},
+	}
+}
